#include "sdkconfig.h"
#include "hwTimer.h"
#include "esp_attr.h"
#include "driver/timer.h"

#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "freertos/queue.h"

#include <math.h>
#include <string.h>

typedef struct {
	tmr_t timer[CPU_LAST_TIMER + 1]; ///< Timer array with needed information about timers
	xQueueHandle queue;			     ///< Alarm queue for deferred callbacks
	TaskHandle_t task;			     ///< Task handle for deferred callbacks
} tmr_driver_t;

// Driver info
static tmr_driver_t *tmr = NULL;

void IRAM_ATTR get_group_idx(int8_t unit, int *groupn, int *idx) {
	switch (unit) {
		case 0: *groupn = 0; *idx = 0; return;
		case 1: *groupn = 0; *idx = 1; return;
		case 2: *groupn = 1; *idx = 0; return;
		case 3: *groupn = 1; *idx = 1; return;
	}
}

static int have_timers(int8_t groupn) {
	int i;

	for(i=0; i < CPU_LAST_TIMER + 1;i++) {
		if ((groupn == 0) && (i > 1)) {
			continue;
		}

		if ((groupn == 1) && (i < 2)) {
			continue;
		}

		if (tmr->timer[i].setup) return 1;
	}

	return 0;
}

static void alarm_task(void *arg) {
	tmr_alarm_t alarm;

    for(;;) {
        xQueueReceive(tmr->queue, &alarm, portMAX_DELAY);

        if (tmr->timer[alarm.unit].callback) {
        	tmr->timer[alarm.unit].callback((void *)((int)alarm.unit));
        }
    }
}

static void IRAM_ATTR isr(void *arg) {
    int unit = (int) arg;
	int groupn, idx;
	portBASE_TYPE high_priority_task_awoken = 0;

	// Get group number / idx
	get_group_idx(unit, &groupn, &idx);

	// Get timer group device
	timg_dev_t *group = (groupn==0?&TIMERG0:&TIMERG1);

	// Check that interrupt is for us
	uint32_t intr_status = group->int_st_timers.val;

    if (intr_status & BIT(idx)) {
    	// Reload alarm value
    	group->hw_timer[idx].update = 1;

    	// Clear inerrupt mask
    	if (idx == 0) {
        	group->int_clr_timers.t0 = 1;
    	} else {
        	group->int_clr_timers.t1 = 1;
    	}

    	// Queue alarm
    	tmr_alarm_t alarm;

    	alarm.unit = unit;

    	xQueueSendFromISR(tmr->queue, &alarm, &high_priority_task_awoken);

    	// Enable alarm again
    	group->hw_timer[idx].config.alarm_en = 1;
    }

    if (high_priority_task_awoken == pdTRUE) {
        portYIELD_FROM_ISR();
    }
}

/*
 * Low-level functions
 *
 */
int tmr_ll_setup(uint8_t unit, uint32_t micros, void(*callback)(void *)) {
	int groupn, idx;

	// Allocate space for driver info
	if (!tmr) {
		tmr = calloc(1, sizeof(tmr_driver_t));
		if (!tmr) {
			return -1;
		}
	}

	if (tmr->timer[unit].setup) {
		return 0;
	}

	// The ISR a message is queued to the queue and callback is
	// executed from a task.
    // Create queue if not created
    if (!tmr->queue) {
        tmr->queue = xQueueCreate(10, sizeof(tmr_alarm_t));
        if (!tmr->queue) {
            return -1;
        }
    }

    // Create task if not created
    if (!tmr->task) {
        BaseType_t xReturn;

        xReturn = xTaskCreatePinnedToCore(alarm_task, "tmral", CONFIG_HWTIMER_THREAD_STACK_SIZE, NULL, CONFIG_HWTIMER_THREAD_PRIORITY, &tmr->task, xPortGetCoreID());
        if (xReturn != pdPASS) {
            return -1;
        }
    }
	

	get_group_idx(unit, &groupn, &idx);

	// Configure time and stop
	// Timer is configure for decrement every 1 usec, so
	// alarm valus is exactly the period, expressed in usec
	timer_config_t config;

	config.alarm_en = 1;
	config.auto_reload = 1;
	config.counter_dir = TIMER_COUNT_UP;
	config.divider = 80;
	config.intr_type = TIMER_INTR_LEVEL;
	config.counter_en = TIMER_PAUSE;

    timer_init(groupn, idx, &config);
    timer_pause(groupn, idx);

    // Set timer at required frequency
    timer_set_counter_value(groupn, idx, 0x00000000ULL);
    timer_set_alarm_value(groupn, idx, micros);

    // Enable timer interrupt
    timer_enable_intr(groupn, idx);
    timer_isr_register(groupn, idx, isr, (void *)((int)unit), ESP_INTR_FLAG_IRAM, &tmr->timer[unit].isrh);

    tmr->timer[unit].setup = 1;
    tmr->timer[unit].callback = callback;

	return 0;
}

void tmr_ll_unsetup(uint8_t unit) {

	if (!tmr->timer[unit].setup) {
		return;
	}

	tmr->timer[unit].callback = NULL;
	tmr->timer[unit].setup = 0;

	// Stop timer
	tmr_ll_stop(unit);

	// Remove interrupt
	esp_intr_free((intr_handle_t)tmr->timer[unit].isrh);

	// If we not have timers destroy queue and task if
	// allocated
	if (!have_timers(-1)) {
		if (tmr->task) {
			vTaskDelete(tmr->task);
		}

		if (tmr->queue) {
			vQueueDelete(tmr->queue);
		}

		free(tmr);
		tmr = NULL;
	}

}

void tmr_ll_start(uint8_t unit) {
	int groupn, idx;
	get_group_idx(unit, &groupn, &idx);
	timer_start(groupn, idx);
}

void tmr_ll_stop(uint8_t unit) {
	int groupn, idx;
	get_group_idx(unit, &groupn, &idx);
	timer_pause(groupn, idx);
}

/*
 * Operation functions
 */
esp_err_t tmr_setup(int8_t unit, uint32_t micros, void(*callback)(void *)) {
	// Sanity checks
	if ((unit < CPU_FIRST_TIMER) || (unit > CPU_LAST_TIMER)) {
		return ESP_ERR_INVALID_ARG;
	}

	if (micros < 5) {
		return ESP_ERR_INVALID_ARG;
	}

	if (tmr_ll_setup(unit, micros, callback) < 0) {
		return ESP_ERR_INVALID_ARG;
	}

	return ESP_OK;
}

esp_err_t tmr_unsetup(int8_t unit) {
	// Sanity checks
	if ((unit < CPU_FIRST_TIMER) || (unit > CPU_LAST_TIMER)) {
		return ESP_ERR_INVALID_ARG;
	}

	tmr_ll_unsetup(unit);

	return ESP_OK;
}

esp_err_t tmr_start(int8_t unit) {
	// Sanity checks
	if ((unit < CPU_FIRST_TIMER) || (unit > CPU_LAST_TIMER)) {
		return ESP_ERR_INVALID_ARG;
	}

	if (!tmr->timer[unit].setup) {
		return ESP_ERR_INVALID_ARG;
	}

	tmr_ll_start(unit);

	return ESP_OK;
}

esp_err_t tmr_stop(int8_t unit) {
	// Sanity checks
	if ((unit < CPU_FIRST_TIMER) || (unit > CPU_LAST_TIMER)) {
		return ESP_ERR_INVALID_ARG;
	}

	if (!tmr->timer[unit].setup) {
		return ESP_ERR_INVALID_ARG;
	}

	tmr_ll_stop(unit);

	return ESP_OK;
}

